import pytest
from src.battlenet_client.client import BattlenetClient
from decouple import config


def test_battlenet_client_client_credentials():
    client = BattlenetClient.client_credential("us", config("CLIENT_ID"), config("CLIENT_SECRET"))
    assert client.client_id == config("CLIENT_ID")
    assert client.client_secret == config("CLIENT_SECRET")
    assert client.region == "us"
    assert client.oauth.access_token is not None
    assert client.oauth.token is not None
    assert isinstance(client.oauth.access_token, str)
    assert isinstance(client.oauth.token, dict)
    assert client.oauth.token["token_type"] == "bearer"
    assert client.oauth.token["expires_in"] == 86399
    assert client.oauth.token["sub"] == config("CLIENT_ID")

def test_battlenet_client_client_credentials_failed_region():
    client = BattlenetClient.client_credential("zz", config("CLIENT_ID"), config("CLIENT_SECRET"))
    assert client.client_id == config("CLIENT_ID")
    assert client.client_secret == config("CLIENT_SECRET")
    assert client.region != "us"
    assert client.oauth.access_token is not None
    assert client.oauth.token is not None
    assert isinstance(client.oauth.access_token, str)
    assert isinstance(client.oauth.token, dict)
    assert client.oauth.token["token_type"] == "bearer"
    assert client.oauth.token["expires_in"] == 86399
    assert client.oauth.token["sub"] == config("CLIENT_ID")


"""Defines the decorators used in this package

Functions:
   verify_region

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
import functools

from .exceptions import BNetGrantError, BNetRegionNotFoundError
from .constants import Region


__version__ = '1.1.0'
__author__ = 'David \'Gahd\' Couples'


def verify_region(func):
    """Verifies the function's first parameter is a valid region abbreviation

    Raises:
        BNetRegionNotFoundError: when the region (arg[0]) is not a valid region tag
    """
    @functools.wraps(func)
    def wrapper_verify_region(*args, **kwargs):
        if args[0].lower() not in Region.VALID_REGIONS:
            raise BNetRegionNotFoundError('Region not found')

        return func(*args, **kwargs)
    return wrapper_verify_region


def verify_grant_type(target_grant_type):
    """ Verfies the grant type is valid

    Args:
        target_grant_type (str): valid grant type

    Raises:
        BNetGrantError: when the grant type is not valid
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if hasattr(self, 'grant_type') and self.grant_type != target_grant_type:
                raise BNetGrantError(f"Invalid grant type: {self.grant_type}")
            return func(self, *args, **kwargs)
        return wrapper

    if callable(target_grant_type):
        return decorator(target_grant_type)
    else:
        return decorator



from typing import Union
import sqlite3

from ..utils.utils import slugify
from .base import BaseCache


class SQLiteCache(BaseCache):

    connector = None

    def __init__(self, connector, region_tag: str, chunk_size: int=1024**3) -> None:
        super().__init__(region_tag, chunk_size)
        self.connector = connector
        self.connector.row_factory = sqlite3.Row

    def __del__(self) -> None:
        self.connector.close()

    def check(self, category_tag: str, *args, **kwargs) -> bool:
        """ Checks if the cache is avaiable for the given arguments

        Args:
            category_tag (str): the category tag
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bool: True if the cache is available and valid for the given arguments else False
        """
        hash_value = self.compute_hash(*args, **kwargs)
        try:
            res = self.connector.execute("""SELECT COUNT(*) FROM cache, categories
                                            WHERE cache.hash = ? AND cache.next_update > unixepoch('now') AND
                                                categories.name = ? AND cache.categoryId = categories.ROWID;""",
                                         (hash_value, slugify(category_tag))).fetchone()
        except TypeError:
            return False
        else:
            return res[0] > 0

    def select(self, category_tag: str, *args, **kwargs) -> Union[None, bytes]:
        """ Retrieves the data from the cache

        Args:
            category_tag (str): category name
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bytes: the data from the cache if available and valid
            None: upon failure or no valid data
        """
        hash_value = self.compute_hash(*args, **kwargs)
        try:
            results = self.connector.execute("""SELECT chunks.data FROM cache, categories, chunks
                                                WHERE cache.hash = ? AND chunks.cacheId = cache.ROWID AND
                                                    categories.name = ? AND cache.categoryId = categories.ROWID
                                                ORDER BY sequence;""", (hash_value, slugify(category_tag))).fetchall()
        except TypeError:
            return None
        else:
            if len(results) == 0:
                return None

            return b''.join([row[0] for row in results])

    def insert(self, category_tag: str, data: bytes, *args, **kwargs) -> bool:
        """ Inserts the data into the cache
        Args:
            category_tag (str): the category tag
            data (bytes): the data to insert
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bool: True if the insertion was successful else False
        """
        hash_value = self.compute_hash(*args, **kwargs)


        try:
            results = self.connector.execute("""INSERT INTO cache (hash, categoryId, next_update) 
                                             SELECT :hash, categories.ROWID, unixepoch(strftime('%Y-%m-%d 10:00:00', 'now', 'weekday 2')) + categories.duration
                                             FROM categories
                                             WHERE categories.name = :name""",
                                             {"hash": hash_value, "name": slugify(category_tag)})
        except sqlite3.Error:
            return False
        else:
            self.connector.commit()
            last_id = results.lastrowid

        chunks = [{"cacheId": last_id, "sequence": counter, "data": chunk} for counter, chunk in enumerate(self.chunk_data(data), start=1)]

        try:
            results = self.connector.executemany("""INSERT INTO chunks (cacheId, sequence, data)
                                                 VALUES (:cacheId, :sequence, :data);""", chunks)
        except sqlite3.Error:
            return False
        else:
            self.connector.commit()
            return results.rowcount > 0


    def update(self, category_tag: str, data: bytes, *args, **kwargs) -> bool:
        """ Insert or update the data in the cache

        Args:
            category_tag (str): category name
            data (bytes): data to insert
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bool: True if the data was inserted or not, False otherwise
        """
        hash_value = self.compute_hash(*args, **kwargs)

        counter = 0
        try:
            cache = self.connector.execute("""SELECT ROWID FROM cache WHERE hash = ?;""", (hash_value,)).fetchone()
        except sqlite3.Error:
            return False

        if not cache:
            return self.insert(category_tag, data, *args, **kwargs)

        chunks = [{'cacheId': cache["ROWID"], "sequence": counter, "data": chunk} for counter, chunk in enumerate(self.chunk_data(data), start=1)]

        try:
            results = self.connector.executemany("""INSERT INTO chunks (cacheId, sequence, data)
                                                    VALUES (:cacheId, :sequence, :data)
                                                    ON CONFLICT (cacheId, sequence) 
                                                        DO UPDATE SET cacheId=excluded.cacheId,
                                                                      sequence=excluded.sequence,
                                                                      data=excluded.data;""", chunks)
        except sqlite3.Error as e:
            print(f"Failed to insert chunk {counter}: {e}")
            return False
        else:
            self.connector.commit()
            return results.lastrowid > 0

    def delete(self, category_tag: str, *args, **kwargs) -> bool:
        """ Deletes the data from the cache

        Args:
            category_tag: category name
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bool: True if the cache is deleted or false if it was not deleted or deleted completely
        """
        hashed_value = self.compute_hash(*args, **kwargs)
        cache_id = self.connector.execute("""SELECT FROM cache WHERE hash = ?;""", (hashed_value,)).fetchone()[0]

        try:
            chunk_results = self.connector.execute("""DELETE FROM chunks WHERE cacheId = ?;""", (cache_id,))
            cache_results = self.connector.execute("""DELETE FROM cache WHERE hash = ?;""", (hashed_value,))
        except TypeError:
            return False
        else:
            self.connector.commit()

        return chunk_results.rowcount and cache_results.rowcount



from abc import ABC, abstractmethod
from hashlib import sha1
from json import dumps


class BaseCache(ABC):
    """ Abstract base class for the cache system

    Atributes:
        region_tag (str): region for the cache
        chunk_size (int): number of bytes to read at a time from the cache
    """

    def __init__(self, region: str, chunk_size=1024**3):
        self.region_tag = region.lower()
        self.chunk_size = chunk_size

    def compute_hash(self, *args, **kwargs) -> str:
        """ Computes the hash of the given arguments

        Args:
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            str: hash of the given arguments
        """
        return sha1(f"{self.region_tag}|{dumps(args)}|{dumps(kwargs)}".encode("utf8")).hexdigest()

    def chunk_data(self, data: bytes) -> list[bytes]:
        """ Breaks the data into chunks

        Args:
            data (bytes): data to chunk

        Returns:
            list of bytes: returns the list of chunks
        """
        if not isinstance(data, bytes):
            raise TypeError("Data must be of type bytes")
        if len(data) < 1:
            raise ValueError("data size must be at least 1 byte")

        chunks = []
        for chunk in range(0, len(data), self.chunk_size):
            chunks.append(data[chunk:chunk+self.chunk_size])
        return chunks

    @abstractmethod
    def check(self, catagory_tag: str, *args, **kwargs) -> bool:
        """ Checks if the cache is avaiable for the given arguments

        Args:
            catagory_tag (str): category name
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bool: True if the cache is available and valid for the given arguments else False

        Raises:
            NotImplementedError: if the function is not implemented in a subclass
        """
        raise NotImplementedError("This method is not implemented")

    @abstractmethod
    def select(self, catagory_tag: str, *args, **kwargs) -> bytes:
        """ Retrieves the data from the cache

        Args:
            catagory_tag (str): category name
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bytes: the data from the cache if available and valid
            None: upon failure or no valid data

        Raises:
            NotImplementedError: if the function is not implemented in a subclass
        """
        raise NotImplementedError("This method is not implemented")

    @abstractmethod
    def insert(self, catagory_tag: str, data: bytes, *args, **kwargs) -> bool:
        """ update the data in the cache

        Args:
            catagory_tag (str): category name
            data (bytes): data to insert
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Raises:
            NotImplementedError: if the function is not implemented in a subclass
        """
        raise NotImplementedError("This method is not implemented")


    @abstractmethod
    def update(self, catagory_tag: str, data: bytes, *args, **kwargs) -> bool:
        """ update the data in the cache

        Args:
            catagory_tag (str): category name
            data (bytes): data to insert
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Raises:
            NotImplementedError: if the function is not implemented in a subclass
        """
        raise NotImplementedError("This method is not implemented")

    @abstractmethod
    def delete(self, catagory_tag: str, *args, **kwargs) -> bool:
        """ Deletes the data from the cache

        Args:
            catagory_tag (str): category name
            *args: list of arguments for the API endpoint
            **kwargs: dict of keyword arguments for the API endpoint

        Returns:
            bool: True if the cache is deleted or false if it was not deleted or deleted completely

        Raises:
            NotImplementedError: if the function is not implemented in a subclass
        """
        raise NotImplementedError("This method is not implemented")
